# Management of dependencies sources
*generated from rpm-build-pyproject/docs/deps.md*

## Overview
This feature allows to manage runtime and/or buildtime dependencies of Python
RPM package in Python-ecosystem manner for the purpose of automation (e.g. such
dependencies can be synced, verified or evaluated). In particular,
<tt><span style="color:green">%pyproject_deps_resync</span></tt>-based macros collect [PEP508](https://peps.python.org/pep-0508/) requirements from different sources and
store them in <tt><span style="color:green">%_sourcedir/pyproject_deps.json</span></tt>
while <tt><span style="color:green">%pyproject_builddeps</span></tt> and <tt><span style="color:green">%pyproject_runtimedeps</span></tt>
evaluate dependencies specified by this config file in current Python environment
and generate <tt><span style="color:green">BuildRequires</span></tt> and <tt><span style="color:green">Requires</span></tt> RPM specfile's
entries respectively. Sources of dependencies can be either well known
standard formats like core metadata, PEP518, PEP517 or tool-specific
formats like pip reqfile, tox deps, etc.

Typical RPM specfile for Python package using this feature looks like:

```RPMSpec
...
Source1: %pyproject_deps_config_name
%pyproject_runtimedeps_metadata
BuildRequires(pre): rpm-build-pyproject
%pyproject_builddeps_build
%if_with check
%pyproject_builddeps_metadata
%endif
...
%prep
%setup
%pyproject_deps_resync_build
%pyproject_deps_resync_metadata
```

<tt><span style="color:green">%pyproject_deps</span></tt>-based RPM macros are shipped in [rpm-build-pyproject](https://packages.altlinux.org/en/sisyphus/srpms/rpm-build-pyproject/)
RPM package, that can be used as drop-in replacement for [rpm-build-python3](https://packages.altlinux.org/en/sisyphus/srpms/rpm-build-python3/) in RPM specfile.

## Bootstrap
How to start using this automation.

- add build dependency(pre) on <tt><span style="color:green">rpm-build-pyproject</span></tt>

    ```spec
    BuildRequires(pre): rpm-build-pyproject
    ```

- add empty configuration for dependencies sources
  - configuration
    ```sh
    echo '{"sources":{}}' > .gear/pyproject_deps.json
    ```

  - gear rules
    ```sh
    echo 'copy: .gear/pyproject_deps.json' >> .gear/rules
    ```

  - RPM spec
    ```spec
    SourceN: %pyproject_deps_config_name
    ```

- add <tt><span style="color:green">%pyproject_builddeps_build</span></tt> and <tt><span style="color:green">%pyproject_deps_resync_build</span></tt> (requires unpacked sources) RPM macros for configuring and syncing [build](#build) dependencies

- add <tt><span style="color:green">%pyproject_runtimedeps_metadata</span></tt> and <tt><span style="color:green">%pyproject_deps_resync_metadata</span></tt> (requires unpacked sources) RPM macros for configuring and syncing [runtime](#runtime) dependencies

- add <tt><span style="color:green">%pyproject_builddeps_metadata</span></tt> and specific check RPM macros for configuring and syncing [tests](#check) dependencies

## BUILD
Build dependencies format is standardized by [PEP518](https://peps.python.org/pep-0518/) and [PEP517](https://peps.python.org/pep-0517/).

- add <tt><span style="color:green">%pyproject_builddeps_build</span></tt> macro into RPM specfile that:

  - eval dependencies of <tt><span style="color:green">PEP518</span></tt> and <tt><span style="color:green">PEP517</span></tt> sources (they will be
    defined later) in Python environment
  - map dependencies to distro names
  - generate RPM <tt><span style="color:green">BuildRequires</span></tt> tags

- append <tt><span style="color:green">%pyproject_deps_resync_build</span></tt> macro to RPM <tt><span style="color:green">%prep</span></tt> section.
  This macro:

  - configure <tt><span style="color:green">PEP518</span></tt> source of dependencies if it doesn't exist yet
  - verify <tt><span style="color:green">PEP518</span></tt> source and resync it if needed
  - configure <tt><span style="color:green">PEP517</span></tt> source of dependencies if it doesn't exist yet
  - verify <tt><span style="color:green">PEP517</span></tt> source and resync it if needed

- run build. The RPM build may fail on verification of <tt><span style="color:green">PEP518</span></tt> dependencies
  (if any). Update stored config with produced one if needed.
  For example,
  ```sh
  hsh-run -- /bin/bash -ec 'cat "$(rpm --eval %pyproject_deps_config)"' > .gear/pyproject_deps.json
  ```

- <tt><span style="color:green">%pyproject_deps</span></tt>-based RPM macros expect to find its configuration in
  RPM sources directory by default.
  Hence,

  - add <tt><span style="color:green">copy: .gear/pyproject_deps.json</span></tt> to .gear/rules
  - add <tt><span style="color:green">Source1: %pyproject_deps_config_name</span></tt> to RPM specfile
    (the source number doesn't matter)

- exclude filters can be applied on evaluation of dependencies.
  For example, filter out dependencies having normalized names that match <tt><span style="color:green">types-</span></tt>:
  ```RPMSpec
  %set_pyproject_deps_build_filter types-
  ```

  or

  ```RPMSpec
  %add_pyproject_deps_build_filter types-
  ```

- rerun build. The RPM build may fail on verification of PEP517 dependencies
  (if any). Update stored config with produced one if needed.
  For example,
  ```sh
  hsh-run -- /bin/bash -ec 'cat "$(rpm --eval %pyproject_deps_config)"' > .gear/pyproject_deps.json
  ```

This is enough to manage *Python* build dependencies of vast of majority of
Python projects.


## RUNTIME

By default, runtime dependencies of Python projects in ALTLinux are generated
automatically by parsing source code and fetching of all of external imports
(this feature is known as <tt><span style="color:green">autoreq</span></tt>). This may or may not work well.
For example:

- conditional or optional dependencies can't be automatically accounted
- installation tools like pip may expect all installed dependencies of projects

Downstream maintainers have to manually handle such cases.

Runtime dependencies format is standardized by [core metadata](https://packaging.python.org/en/latest/specifications/core-metadata/) and it's usually
generated by build backend from some tool-specific format. Produced list of
dependencies is not *necessary and sufficient*. But it may complement
auto-generated dependencies to cover unhandled cases.

- add <tt><span style="color:green">%pyproject_runtimedeps_metadata</span></tt> macro into RPM specfile that:

  - eval dependencies of <tt><span style="color:green">metadata</span></tt> source (it will be defined later)
    in Python environment
  - map dependencies to distro names
  - generate RPM <tt><span style="color:green">Requires</span></tt> tags

Note: if some specific [extras](https://peps.python.org/pep-0508/#extras) is required the <tt><span style="color:green">%pyproject_runtimedeps_metadata_extra</span></tt> macro can be used instead (in this case all the runtime dependencies of
current package **and** dependencies that are marked with the given extra of
current package will be installed)

- append <tt><span style="color:green">%pyproject_deps_resync_metadata</span></tt> macro to RPM <tt><span style="color:green">%prep</span></tt> section.
  This macro:

  - configure <tt><span style="color:green">core metadata</span></tt> source of dependencies if it doesn't exist yet
  - verify <tt><span style="color:green">core metadata</span></tt> source and resync it if needed

- rerun build. The RPM build may fail on verification of <tt><span style="color:green">metadata</span></tt>
  dependencies (if any). Update stored config with produced one if needed.
  For example,
  ```sh
  hsh-run -- /bin/bash -ec 'cat "$(rpm --eval %pyproject_deps_config)"' > .gear/pyproject_deps.json
  ```

- exclude filters can be applied on evaluation of dependencies:
  For example, filter out dependencies having normalized names that match <tt><span style="color:green">types-</span></tt>:
  ```RPMSpec
  %set_pyproject_deps_runtime_filter types-
  ```

  or

  ```RPMSpec
  %add_pyproject_deps_runtime_filter types-
  ```

## CHECK
Tests dependencies format is either *standardized* with:

- [core metadata](#core-metadata)
- [dependency groups](#dependency-group)

or *tool-specific*. But most *popular* of them are supported by
<tt><span style="color:green">pyproject_installer</span></tt>:

- [pip's reqfile](#pip-reqfile)
- [poetry group](#poetry-group)
- [tox env](#tox-env)
- [hatch env](#hatch-env)
- [pdm group](#pdm-group)
- [pipenv category](#pipenv-category)

Usually tests require all [runtime](#core-metadata) dependencies of project *and* additional
packages for test environment (e.g. pytest). Thus, configuration of several
sources may be required (e.g. metadata + pip_reqfile).

### check filter
There can be many of redundant dependencies like:

- <tt><span style="color:red">code style</span></tt> checkers that are really helpful in development but they
  often report different results from version to version.
- <tt><span style="color:red">coverage</span></tt> measurement, it's pointless overhead to do this for RPM
  packaging.
- <tt><span style="color:red">publishing tools</span></tt>
- <tt><span style="color:red">documentation tools</span></tt>

Hence, there is a predefined filter of <tt><span style="color:green">check</span></tt> dependencies
<tt><span style="color:green">%pyproject_deps_check_filter</span></tt>. This filter is applied automatically on
<tt><span style="color:green">%pyproject_builddeps_check</span></tt> or <tt><span style="color:green">%pyproject_builddeps_metadata</span></tt> macro usage.

Today's default filter includes:

- <tt><span style="color:green">%{?pypi_name:'%{pep503_name %pypi_name}$'}</span></tt>
- flake8 and its plugins
- black and its plugins
- coverage and its plugins
- codecov and its plugins
- coveralls and its plugins
- mypy and its plugins
- isort and its plugins
- pre-commit and its plugins
- pytest-checkdocs
- twine and its plugins
- ruff and its plugins
- typing stubs
- pyright and its plugins
- Sphinx and its plugins

This filter can be changed like:
```RPMSpec
%set_pyproject_deps_check_filter foo bar
```

or

```RPMSpec
%add_pyproject_deps_check_filter foo bar
```

Note: this is exposed as *deps* <tt><span style="color:green">--exclude</span></tt> option that accepts regex
patterns and excludes requirement having [PEP503-normalized name](https://peps.python.org/pep-0503/#normalized-names) which
matches one of these patterns. Thus, if only the name should be excluded then
this name should be normalized first:
```RPMSpec
%add_pyproject_deps_check_filter '%{pep503_name Foo-Project.my}$'
```

### core metadata
Specification:

- https://packaging.python.org/en/latest/specifications/core-metadata/
- https://packaging.python.org/en/latest/specifications/core-metadata/#requires-dist-multiple-use

For example, upstream use core metadata as source of tests dependencies and
mark them with <tt><span style="color:green">testing</span></tt> extra (this means that all runtime dependencies of
current package and any dependencies that are listed in the <tt><span style="color:green">testing</span></tt> extra of
current package will be installed, see [extras](https://peps.python.org/pep-0508/#extras)). Usually this is exposed as <tt><span style="color:green">extras = testing</span></tt> in <tt><span style="color:green">tox.ini</span></tt> or
<tt><span style="color:green">pip install .[testing]</span></tt> in CI config.

- add macro to RPM specfile
  ```RPMSpec
  %if_with check
  %pyproject_builddeps_metadata_extra testing
  %endif
  ```
  This macro:

  - eval dependencies of <tt><span style="color:green">metadata</span></tt> source (it's already defined as source of
    [runtime](#runtime) dependencies, otherwise it should be defined manually) in Python
    environment with <tt><span style="color:green">testing</span></tt> extra
  - filter out dependencies according to <tt><span style="color:green">%pyproject_deps_check_filter</span></tt>
  - map dependencies to distro names
  - generate RPM <tt><span style="color:green">BuildRequires</span></tt> tags

- filter out extra dependencies with [check-filter](#check-filter)

### pip reqfile
Specification:

- https://pip.pypa.io/en/stable/reference/requirements-file-format/#requirements-file-format

Limitations:

- supported only [PEP508](https://peps.python.org/pep-0508/) requirements
- line continuations are not supported
- pip's inline options are not supported

For example, upstream use pip's reqfile as source of tests dependencies and
it's saved as <tt><span style="color:green">test-requirements.txt</span></tt>.

- add macro to RPM specfile
  ```RPMSpec
  %if_with check
  %pyproject_builddeps_check
  %endif
  ```
  This macro:

  - eval dependencies of <tt><span style="color:green">check</span></tt> source (it will be defined later) in
    Python environment
  - filter out dependencies according to <tt><span style="color:green">%pyproject_deps_check_filter</span></tt>
  - map dependencies to distro names
  - generate RPM <tt><span style="color:green">BuildRequires</span></tt> tags

- configure <tt><span style="color:green">check</span></tt> source of dependencies by appending to RPM <tt><span style="color:green">%prep</span></tt> section:
  ```RPMSpec
  %if_with check
  %pyproject_deps_resync_check_pipreqfile test-requirements.txt
  %endif
  ```

  or source having arbitrary name

  ```RPMSpec
  %if_with check
  %pyproject_deps_resync check pip_reqfile test-requirements.txt
  %endif
  ```

- rerun build. The RPM build may fail on verification of <tt><span style="color:green">check</span></tt> dependencies
  (if any). Update stored config with produced one if needed.
  For example,
  ```sh
  hsh-run -- /bin/bash -ec 'cat "$(rpm --eval %pyproject_deps_config)"' > .gear/pyproject_deps.json
  ```

- filter out extra dependencies with [check-filter](#check-filter)

### poetry group
Specification:

-  https://python-poetry.org/docs/dependency-specification/
-  https://python-poetry.org/docs/managing-dependencies/

Limitations:

- versions are not supported
  (poetry version format is not compatible with [PEP440](https://peps.python.org/pep-0440/))
- <tt><span style="color:red">python</span></tt> field is not supported

For example, upstream uses poetry dependencies group <tt><span style="color:green">dev</span></tt> as source of tests
dependencies. By default, poetry installs dependencies of all non-optional
groups including dependencies of project. Hence, [runtime](#runtime) should be
configured first.

- add macro to RPM specfile
  ```RPMSpec
  %if_with check
  %pyproject_builddeps_metadata
  %pyproject_builddeps_check
  %endif
  ```
  These macros:

  - eval dependencies of <tt><span style="color:green">metadata</span></tt> and <tt><span style="color:green">check</span></tt> sources
    (<tt><span style="color:green">metadata</span></tt> is already defined as source of [runtime](#runtime) dependencies,
    otherwise it should be defined manually) in Python environment
  - filter out dependencies according to <tt><span style="color:green">%pyproject_deps_check_filter</span></tt>
  - map dependencies to distro names
  - generate RPM <tt><span style="color:green">BuildRequires</span></tt> tags

- configure <tt><span style="color:green">check</span></tt> source of dependencies by appending to RPM <tt><span style="color:green">%prep</span></tt> section:
  ```RPMSpec
  %if_with check
  %pyproject_deps_resync_check_poetry dev
  %endif
  ```

  or source having arbitrary name

  ```RPMSpec
  %if_with check
  %pyproject_deps_resync check poetry dev
  %endif
  ```

- rerun build. The RPM build may fail on verification of <tt><span style="color:green">check</span></tt> dependencies
  (if any). Update stored config with produced one if needed.
  For example,
  ```sh
  hsh-run -- /bin/bash -ec 'cat "$(rpm --eval %pyproject_deps_config)"' > .gear/pyproject_deps.json
  ```

- filter out extra dependencies with [check-filter](#check-filter)

### tox env
Specification:

- https://tox.wiki/en/latest/config.html#deps
- config formats:
  https://tox.wiki/en/latest/config.html#discovery-and-file-types

Limitations:

- supported only [PEP508](https://peps.python.org/pep-0508/) requirements
- config substitutions are not supported
- <tt><span style="color:green">testenv</span></tt> inheritance is not supported
- pip's req files or options are not supported

Note: if <tt><span style="color:green">deps</span></tt> includes pip's requirements file then use [pip's reqfile](#pip-reqfile) instead. If <tt><span style="color:green">testenv</span></tt> has <tt><span style="color:green">extras</span></tt> then use [core metadata](#core-metadata) instead.

For example, upstream uses tox' <tt><span style="color:green">deps</span></tt> of <tt><span style="color:green">testenv</span></tt> section of <tt><span style="color:green">tox.ini</span></tt> config
as source of tests dependencies. By default, <tt><span style="color:green">tox</span></tt> installs these
dependencies prior to package installation and then installs package and its
runtime dependencies. Hence, [runtime](#runtime) should be configured first.

- add macro to RPM specfile
  ```RPMSpec
  %if_with check
  %pyproject_builddeps_metadata
  %pyproject_builddeps_check
  %endif
  ```
  These macros:

  - eval dependencies of <tt><span style="color:green">metadata</span></tt> and <tt><span style="color:green">check</span></tt> sources
    (<tt><span style="color:green">metadata</span></tt> is already defined as source of [runtime](#runtime) dependencies,
    otherwise it should be defined manually) in Python environment
  - filter out dependencies according to <tt><span style="color:green">%pyproject_deps_check_filter</span></tt>
  - map dependencies to distro names
  - generate RPM <tt><span style="color:green">BuildRequires</span></tt> tags

- configure <tt><span style="color:green">check</span></tt> source of dependencies by appending to RPM <tt><span style="color:green">%prep</span></tt> section:
  ```RPMSpec
  %if_with check
  %pyproject_deps_resync_check_tox tox.ini testenv
  %endif
  ```

  or source having arbitrary name

  ```RPMSpec
  %if_with check
  %pyproject_deps_resync check tox tox.ini testenv
  %endif
  ```

- rerun build. The RPM build may fail on verification of <tt><span style="color:green">check</span></tt> dependencies
  (if any). Update stored config with produced one if needed.
  For example,
  ```sh
  hsh-run -- /bin/bash -ec 'cat "$(rpm --eval %pyproject_deps_config)"' > .gear/pyproject_deps.json
  ```

- filter out extra dependencies with [check-filter](#check-filter)

### hatch env
Specification:

- https://hatch.pypa.io/latest/config/environment/overview/#dependencies
- https://hatch.pypa.io/latest/intro/#configuration

Limitations:

- supported only [PEP508](https://peps.python.org/pep-0508/) requirements
- env inheritance is not supported
- context formatting is not supported
- features(extra) is not supported

Note: if <tt><span style="color:green">env</span></tt> has <tt><span style="color:green">features</span></tt> then use [core metadata](#core-metadata) with extra instead.

For example, upstream uses hatch's <tt><span style="color:green">dependencies</span></tt> of <tt><span style="color:green">test</span></tt> env in <tt><span style="color:green">hatch.toml</span></tt> config
as source of tests dependencies. By default, <tt><span style="color:green">hatch</span></tt> installs these
dependencies in addition to the runtime ones. Hence, [runtime](#runtime) should be configured first.

- add macro to RPM specfile
  ```RPMSpec
  %if_with check
  %pyproject_builddeps_metadata
  %pyproject_builddeps_check
  %endif
  ```
  These macros:

  - eval dependencies of <tt><span style="color:green">metadata</span></tt> and <tt><span style="color:green">check</span></tt> sources
    (<tt><span style="color:green">metadata</span></tt> is already defined as source of [runtime](#runtime) dependencies,
    otherwise it should be defined manually) in Python environment
  - filter out dependencies according to <tt><span style="color:green">%pyproject_deps_check_filter</span></tt>
  - map dependencies to distro names
  - generate RPM <tt><span style="color:green">BuildRequires</span></tt> tags

- configure <tt><span style="color:green">check</span></tt> source of dependencies by appending to RPM <tt><span style="color:green">%prep</span></tt> section:
  ```RPMSpec
  %if_with check
  %pyproject_deps_resync_check_hatch hatch.toml test
  %endif
  ```

  or source having arbitrary name

  ```RPMSpec
  %if_with check
  %pyproject_deps_resync YOURSOURCENAME hatch hatch.toml test
  %endif
  ```

- rerun build. The RPM build may fail on verification of <tt><span style="color:green">check</span></tt> dependencies
  (if any). Update stored config with produced one if needed.
  For example,
  ```sh
  hsh-run -- /bin/bash -ec 'cat "$(rpm --eval %pyproject_deps_config)"' > .gear/pyproject_deps.json
  ```

- filter out extra dependencies with [check-filter](#check-filter)

### pdm group
Specification:

-  https://pdm.fming.dev/latest/usage/dependency/#add-development-only-dependencies

Limitations:

- supported only [PEP508](https://peps.python.org/pep-0508/) requirements

For example, upstream uses pdm dev dependencies' group <tt><span style="color:green">test</span></tt> as source of tests
dependencies. By default, pdm installs dependencies of all non-optional
groups including runtime dependencies of project. Hence, [runtime](#runtime) should be
configured first.

- add macro to RPM specfile
  ```RPMSpec
  %if_with check
  %pyproject_builddeps_metadata
  %pyproject_builddeps_check
  %endif
  ```
  These macros:

  - eval dependencies of <tt><span style="color:green">metadata</span></tt> and <tt><span style="color:green">check</span></tt> sources
    (<tt><span style="color:green">metadata</span></tt> is already defined as source of [runtime](#runtime) dependencies,
    otherwise it should be defined manually) in Python environment
  - filter out dependencies according to <tt><span style="color:green">%pyproject_deps_check_filter</span></tt>
  - map dependencies to distro names
  - generate RPM <tt><span style="color:green">BuildRequires</span></tt> tags

- configure <tt><span style="color:green">check</span></tt> source of dependencies by appending to RPM <tt><span style="color:green">%prep</span></tt> section:
  ```RPMSpec
  %if_with check
  %pyproject_deps_resync_check_pdm test
  %endif
  ```

  or source having arbitrary name

  ```RPMSpec
  %if_with check
  %pyproject_deps_resync YOURSOURCENAME pdm test
  %endif
  ```

- rerun build. The RPM build may fail on verification of <tt><span style="color:green">check</span></tt> dependencies
  (if any). Update stored config with produced one if needed.
  For example,
  ```sh
  hsh-run -- /bin/bash -ec 'cat "$(rpm --eval %pyproject_deps_config)"' > .gear/pyproject_deps.json
  ```

- filter out extra dependencies with [check-filter](#check-filter)

### pipenv category
Specification:

- https://pipenv.pypa.io/en/latest/pipfile.html
- https://pipenv.pypa.io/en/latest/specifiers.html

Limitations:

- standalone [PEP508 specifiers](https://peps.python.org/pep-0508/#environment-markers) as keys are not supported

For example, upstream uses pipenv dev dependencies' category <tt><span style="color:green">dev-packages</span></tt> as a source of tests
dependencies. By default, pipenv installs dependencies of <tt><span style="color:green">packages</span></tt>
category and [runtime dependencies](#runtime) should be
configured first.

- add macro to RPM specfile
  ```RPMSpec
  %if_with check
  %pyproject_builddeps_metadata
  %pyproject_builddeps_check
  %endif
  ```
  These macros:

  - eval dependencies of <tt><span style="color:green">metadata</span></tt> and <tt><span style="color:green">check</span></tt> sources
    (<tt><span style="color:green">metadata</span></tt> is already defined as source of [runtime](#runtime) dependencies,
    otherwise it should be defined manually) in Python environment
  - filter out dependencies according to <tt><span style="color:green">%pyproject_deps_check_filter</span></tt>
  - map dependencies to distro names
  - generate RPM <tt><span style="color:green">BuildRequires</span></tt> tags

- configure <tt><span style="color:green">check</span></tt> source of dependencies by appending to RPM <tt><span style="color:green">%prep</span></tt> section:
  ```RPMSpec
  %if_with check
  %pyproject_deps_resync_check_pipenv Pipfile dev-packages
  %endif
  ```

  or source having arbitrary name

  ```RPMSpec
  %if_with check
  %pyproject_deps_resync YOURSOURCENAME pipenv Pipfile dev-packages
  %endif
  ```

- rerun build. The RPM build may fail on verification of <tt><span style="color:green">check</span></tt> dependencies
  (if any). Update stored config with produced one if needed.
  For example,
  ```sh
  hsh-run -- /bin/bash -ec 'cat "$(rpm --eval %pyproject_deps_config)"' > .gear/pyproject_deps.json
  ```

- filter out extra dependencies with [check-filter](#check-filter)

### dependency group
Specification:

- https://peps.python.org/pep-0735/#specification

For example, upstream uses <tt><span style="color:green">test</span></tt> dependency group as a source of tests
dependencies. Installation of a dependency group doesn't imply installation of a package's dependencies or the package itself. Thereby [runtime](#runtime) dependencies should be configured first if needed.

- add macro to RPM specfile
  ```RPMSpec
  %if_with check
  %pyproject_builddeps_metadata
  %pyproject_builddeps_check
  %endif
  ```
  These macros:

  - eval dependencies of <tt><span style="color:green">metadata</span></tt> and <tt><span style="color:green">check</span></tt> sources
    (<tt><span style="color:green">metadata</span></tt> is already defined as source of [runtime](#runtime) dependencies,
    otherwise it should be defined manually) in Python environment
  - filter out dependencies according to <tt><span style="color:green">%pyproject_deps_check_filter</span></tt>
  - map dependencies to distro names
  - generate RPM <tt><span style="color:green">BuildRequires</span></tt> tags

- configure <tt><span style="color:green">check</span></tt> source of dependencies by appending to RPM <tt><span style="color:green">%prep</span></tt> section:
  ```RPMSpec
  %if_with check
  %pyproject_deps_resync_check_depgroup test
  %endif
  ```

  or source having arbitrary name

  ```RPMSpec
  %if_with check
  %pyproject_deps_resync YOURSOURCENAME pep735 test
  %endif
  ```

- rerun build. The RPM build may fail on verification of <tt><span style="color:green">check</span></tt> dependencies
  (if any). Update stored config with produced one if needed.
  For example,
  ```sh
  hsh-run -- /bin/bash -ec 'cat "$(rpm --eval %pyproject_deps_config)"' > .gear/pyproject_deps.json
  ```

- filter out extra dependencies with [check-filter](#check-filter)


## Mapping project names to distro names
Python ecosystem uses PEP503 normalization rules for names comparison:

- https://packaging.python.org/en/latest/specifications/name-normalization/#package-name-normalization
- https://peps.python.org/pep-0503/#normalized-names

Mapping rule in ALTLinux:
```
project_name => python3-module-%{pep503_name project_name}
```

This almost always works out of the box, but distro packages having
unnormalized names *must* provide such the name:
```
Provides: python3-module-%{pep503_name %pypi_name} = %EVR
```

The following naming rule for new Python distro packages is strongly encouraged:
```
Name: python3-module-%{pypi_nname}
```
where <tt><span style="color:green">%pypi_nname</span></tt> must be equal to the result of evaluation of
<tt><span style="color:green">%{pep503_name %pypi_name}</span></tt> macro.

## Disable verify mode on syncing deps
By default <tt><span style="color:green">%pyproject_deps_resync_*</span></tt> macros sync dependencies,
save updated configuration, but print diff and exit with code 4 if deps were
unsynced.

It can be helpful for automation or bootstrap to make it possible to disable
verify mode (to not fail on unsynced deps).

One can do this by defining <tt><span style="color:green">pyproject_deps_verify_skip</span></tt> macro name,
e.g. in RPM specfile:
```spec
%define pyproject_deps_verify_skip 1
```

or via rpmbuild's cli:
```sh
rpmbuild --define='pyproject_deps_verify_skip 1' ...
```
